/*
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
 * SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Mir support
#include <mir/run_mir.h>
#include <mir/abnormal_exit.h>
#include <mir/default_server_configuration.h>
#include <mir/display_server.h>
#include <mir/main_loop.h>

// Platform API
#include <ubuntu/application/init.h>

// Qt
#include <QCoreApplication>
#include <QDebug>
#include <QThread>
#include <QFile>

// Std
#include <chrono>
#include <csignal>
#include <thread>

// local
#include "logging.h"
#include "qmirserver.h"

QMirServer::QMirServer(int argc, const char* argv[], QObject *parent)
    : QObject(parent)
    , m_argc(argc)
    , m_argv(argv)
{
}

QMirServer::~QMirServer()
{
}

int QMirServer::runWithClient(std::function<int(int, const char**, ShellServerConfiguration*)> client)
{
    // need local copies as lambda function will not capture class members
    int argc = m_argc;
    auto argv = m_argv;
    auto config = new ShellServerConfiguration(m_argc, m_argv);
    std::thread *t;
    bool mirServerShutDown = false;

    QFile file("/etc/ubuntu-platform-api/application.conf");
    if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QByteArray plugin = file.readLine();
        if (plugin.contains("mirclient")) {
            plugin.replace("mirclient", QByteArray("mirserver"));
            plugin = plugin.trimmed();
            setenv("UBUNTU_PLATFORM_API_BACKEND", plugin.data(), 1);
        }
    }

    mir::run_mir(*config, [&](mir::DisplayServer &server) {
        u_application_init(dynamic_cast<mir::DefaultServerConfiguration*>(config));

        // Mir initiates shutdown on SIGINT & SIGTERM, need to distinguish that shutdown from a
        // client-initiated shutdown. Do this by installing a custom signal handler that is run
        // after Mir's initiate-shutdown handler is called.
        config->the_main_loop()->register_signal_handler(
        {SIGINT, SIGTERM},
        [&](int)
        {
            DLOG("Mir-initiated shut down in progress");
            mirServerShutDown = true;
        });

        try {
            t = new std::thread([&]() {
                client (argc, argv, config);
                if (!mirServerShutDown) {
                    DLOG("Qt-initiated shut down in progress");
                    server.stop();
                }
            });
        } catch (...) {
            LOG("Exception caught, quitting");
        }
    });

    if (QCoreApplication::instance()) {
        bool aboutToQuitSignaled = false;
        QObject::connect(QCoreApplication::instance(), &QCoreApplication::aboutToQuit, [&]() {
                aboutToQuitSignaled = true;
            }
        );
        QCoreApplication::quit();
        while (!aboutToQuitSignaled) {
            std::chrono::milliseconds duration(100);
            std::this_thread::sleep_for(duration);
            QCoreApplication::quit();
        }
    }
    t->join();
    u_application_finish();
    delete config;
    return 0;
}

// class factory implementation
extern "C" {
    QMirServer *createQMirServer(int argc, const char **argv) {
        return new QMirServer(argc, argv);
    }

    int runQMirServerWithClient(QMirServer *mirServer, std::function<int(int, const char**, ShellServerConfiguration*)> client) {
        return mirServer->runWithClient(client);
    }

    void destroyQMirServer(QMirServer *mirServer) {
        delete mirServer;
    }
}
